/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Properties;

import com.ibm.rqm.adapter.library.CommandLineParser;
import com.ibm.rqm.adapter.library.ConnectionInfo;
import com.ibm.rqm.adapter.library.connection.Base64;
import com.ibm.rqm.adapter.library.data.AbstractProcessor;
import com.ibm.rqm.adapter.library.data.AdapterInformation;
import com.ibm.rqm.adapter.library.data.NewRequester;
import com.ibm.rqm.adapter.library.data.RQMAdapter;
import com.ibm.rqm.adapter.library.data.ServerInformation;

/**
 * The entry point to the adapter. This is the class which is instantiated to
 * start up the adapter. It keeps all the information about the adapter instance
 * and delegates the task of communicating to the QM server via
 * {@link NewRequester} and the adapter specific implementation of
 * {@link AbstractProcessor} (for Command Line adapter its
 * {@link QMCommandLineProcessor})
 * 
 * <br>
 * Once the adapter is instantiated, it starts polling the server for new
 * instructions via NewRequester {@link RQMAdapter#start()}
 * 
 * If an instruction is fetched the adapter then delegates the servicing of the
 * instruction to the processor {@link QMCommandLineProcessor}.
 * 
 * @author amathur
 * 
 */
public class QMCommandLineAdapter extends RQMAdapter {

	/**
	 * <code>ADAPTER_TYPE</code> The type which is registered with the QM
	 * server via extension point
	 * <code>com.ibm.rqm.planning.common.executionScriptType</code> attribute
	 * <code>id</code>
	 */
	private static final String ADAPTER_TYPE = "com.ibm.rqm.adapter.commandline"; //$NON-NLS-1$
	private static final String COMMENT = "RQM Command Line Configuration File"; //$NON-NLS-1$
	private static final String DEFAULT_ADAPTER_DESCRIPTION = "Its a command Line Adapter";
	private static final String CLA_ADAPTER_DEFAULT_NAME = "Commandline Adapter";
	private static final String DEFAULT_PREFERENCES_NAME = "commandline.properties";
	
	/**
	 * @param connectionInfo
	 * @param adapterType
	 * @param adapterName
	 * @param sleepTime
	 * @param verbose
	 */
	public QMCommandLineAdapter(ServerInformation server, AdapterInformation adapter) {
		super(server, adapter);
		
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see com.ibm.rqm.adapter.library.data.RQMAdapter#createProcessor(com.ibm.rqm.adapter.library.data.NewRequester)
	 */
	protected AbstractProcessor createProcessor(NewRequester requester) {
		return new QMCommandLineProcessor(requester);
	}

	/**
	 * The supported arguments are
	 * {@link CommandLineParser#ADAPTER_ID_ARGUMENT_NAME},
	 * {@link CommandLineParser#PASSWORD_ARGUMENT_NAME},
	 * {@link CommandLineParser#USER_ID_ARGUMENT_NAME},
	 * {@link CommandLineParser#SLEEP_TIME_ARGUMENT_NAME},
	 * {@link CommandLineParser#REPOSITORY_ARGUMENT_NAME},
	 * {@link CommandLineParser#LOG_FILE_ARGUMENT_NAME}
	 * {@link CommandLineParser#PREFERENCES_ARGUMENT_NAME}
	 * 
	 * However, the individual adapters might want to default a few arguments to make the launch command simpler
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		//load arguments into properties
		Properties arguments = new Properties();
		for (int i = 0; i < args.length; i = i + 2) {
			if ( args.length > i + 1) {
				arguments.put(args[i], args[i+1]);
			}
		}
		
		boolean firstStart = true;
		File configFile = null;
		if(arguments.containsKey(CommandLineParser.CONFIG_FILE_PATH)){//parameterExists(args, CommandLineParser.CONFIG_FILE_PATH)){
			String filePath = arguments.getProperty(CommandLineParser.CONFIG_FILE_PATH);//parameterValue(args, CommandLineParser.CONFIG_FILE_PATH);
			if(filePath != null){
				configFile = new File(filePath);
			}
			else
			{
				Logger.Log.error("Invalid arguments");
				System.exit(1);
			}
		}
		else{
			configFile = new File(CommandLineParser.CONFIG_FILE_NAME);
		}

		// First load the connection info object
		ConnectionInfo conn = CommandLineParser.loadArguments(args);
		
		if (conn == null) {
			Logger.Log.error("Invalid arguments");
			System.exit(1);
		}
		
		try {
                    if (configFile != null && configFile.isDirectory()){
                            throw new IllegalArgumentException(" -configFile accepts the path to a configuration file. The value passed is a directory.");
                    }
                    if ((configFile != null) && (configFile.exists()) ) {
                                    firstStart = false;
                                    Logger.Log.info("Adapter is reconnecting. Using saved configuration information from " + configFile.getName());
                    }
                    if(configFile != null && !configFile.exists()){
                            Logger.Log.warn("Configuration file " + configFile.getAbsolutePath() +" does not exist, adapter will use the arguments passed");
                    }
                }
                catch (Exception e) {
                        Logger.Log.error("QMCommandLineAdapter : Error while starting");
                        Logger.Log.error(e.getMessage());
                        Logger.Log.debug(e);
                        System.exit(1);
                }
		
		// Use Command line adapter's default name
		if (conn.getAdapterName().equals(CommandLineParser.DEFAULT_ADAPTER_NAME)) {
			conn.setAdapterName(CLA_ADAPTER_DEFAULT_NAME);
		}
		
		String password = conn.getPassword();
		if (!firstStart && !(arguments.containsKey(CommandLineParser.PASSWORD_ARGUMENT_NAME)/*parameterExists(args, CommandLineParser.PASSWORD_ARGUMENT_NAME)*/)) {
			try {
				password = new String((new Base64()).decode(conn.getPassword().getBytes("UTF-8")), "UTF-8");
			} catch (UnsupportedEncodingException e1) {
				Logger.Log.error("QMCommandLineAdapter : password decode error");
				Logger.Log.debug(e1);
			}
		}
		
		ServerInformation server = new ServerInformation(conn.getRepositoryAddress());
		server.setLogin(conn.getUserId());
		server.setPassword(password);
		server.setProjectAreaAlias(conn.getProjectAreaAlias());
		
		//Set Proxy info under serverInformation instance
		server.setProxy(conn.getProxy());
		server.setProxyPort(conn.getProxyPort());
		server.setProxyUser(conn.getProxyUser());
		server.setProxyPassword(conn.getProxyPassword());
		AdapterInformation adapter = new AdapterInformation(conn.getAdapterId(), conn.getAdapterName(), ADAPTER_TYPE);
		adapter.setDescription(DEFAULT_ADAPTER_DESCRIPTION);
		adapter.setPollingInterval((int)conn.getPollingInterval());
		
		// This will login the adapter with the QM Server
		QMCommandLineAdapter commandLine = new QMCommandLineAdapter(server, adapter);
		
		// check if server registration exists
		boolean isRegistered = false;
		if (!firstStart){
			try {
				isRegistered = commandLine.getRequester().isAdapterRegistered();
				if (!isRegistered){
					Logger.Log.info("Stored configuration is out of sync. Adapter is re-registering.");
				}
			} catch (Exception e){
				isRegistered = false;
			}
		}
		
		// If its the first start, register and store properties into a config file
		Properties configProps = loadConfigFile(configFile);
		boolean regArgsChanged = isRegisteredArgsChanged(arguments, configProps);
		if (firstStart || regArgsChanged || !isRegistered) {
			try {
				// Register from here so that I can store the data
				commandLine.getRequester().register();
				
				// Now store the information in a config file for future re-use
				if (firstStart || regArgsChanged){
					storeProperties(server, adapter, configFile);
				}
			} catch (Exception e) {
				Logger.Log.warn("Failed to save configuration file for the command line adapter");
				Logger.Log.debug(e);
			}
		}
		else {
			commandLine.getRequester().getAdapter().setRegistered(true);
			storeProperties(server, adapter, configFile);
		}

        if (!commandLine.requester.wasInitialLoginSuccessful()) {
            Logger.Log.error("Failed to log in to server at " + conn.getRepositoryAddress() + 
            		". Please check the repository URL " + conn.getRepositoryAddress() + 
            		", user name " + conn.getUserId() +  
            		" and its password and try again. Make sure that the Server is available.");
            System.exit(1);
        }
        
        File prefsFile = null;
        try { // load preferences, used default file if explicit file not specified
        	prefsFile = ( arguments.containsKey(CommandLineParser.PREFERENCES_ARGUMENT_NAME) ?
        			      new File((String)arguments.get(CommandLineParser.PREFERENCES_ARGUMENT_NAME)) :
        				  new File(DEFAULT_PREFERENCES_NAME) );
        	new CommandLinePreferences(prefsFile);
        } catch ( Exception e ) {
        	Logger.Log.warn("Preferences file can not be loaded : "+prefsFile);
        	Logger.Log.debug(e);
        }

        Logger.Log.info("Created Command Line Adapter");
		Thread thread = commandLine.start();

		try {
			thread.join();
		} catch (InterruptedException exception) {
			// Ignore. The thread is done, that's all we care about.
		}
	}

//	public static boolean parameterExists(String[] args,String parameterName)
//	{
//		 for (int i = 0; i < args.length; i++) {
//	       String argumentName = args[i];
//	       if (argumentName.equalsIgnoreCase(parameterName)) {
//	         	return true;
//	       }
//		 }
//		return false;
//	}
//	
//	public static String parameterValue(String[] args, String argName)
//	{
//		 for (int i = 0; i < args.length; i++) {
//	            String argumentName = args[i];
//	            if (argumentName.equalsIgnoreCase(argName)) {
//	                if (args.length > i + 1) {
//	                    return args[i + 1];
//	                }
//	            }
//		 }
//		 return null;
//	}
//	
//	public static String propertyValue(String key, File configFile)
//	{
//		try {
//			if(configFile.exists()){
//				Properties properties = new Properties();
//				properties.load(new FileInputStream(configFile));
//				return properties.getProperty(key);
//			}
//			else{
//				Logger.Log.error("Configuration file:" + configFile.getName() + "not found, while trying to get property with key:"+ key);
//				return null;
//			}
//		} catch (Exception e) {
//			Logger.Log.error("Property with key:"+ key + " does not exist in configuration file" + configFile.getName(), e);
//			return null;
//		} 
//	}
	
	public static void storeProperties(ServerInformation server, AdapterInformation adapter, File configFile)
	{
		FileOutputStream fos = null;
		try{
			Properties properties = new Properties();
			properties.setProperty(CommandLineParser.REPOSITORY_PROPERTY_NAME, server.getUrl());
			properties.setProperty(CommandLineParser.ADAPTER_ID_PROPERTY_NAME, adapter.getId());
			properties.setProperty(CommandLineParser.USER_ID_PROPERTY_NAME, server.getLogin());
			properties.setProperty(CommandLineParser.PASSWORD_PROPERTY_NAME, (new Base64()).encode(server.getPassword().getBytes("UTF-8")));
			properties.setProperty(CommandLineParser.SLEEP_TIME_PROPERTY_NAME, Integer.toString(adapter.getPollingInterval()));
			properties.setProperty(CommandLineParser.ADAPTER_PROPERTY_NAME, adapter.getName());
			if(null != server.getProjectAreaAlias())//If default project area this will be null
				properties.setProperty(CommandLineParser.PROJECTAREA_ALIAS_PROPERTY_NAME, server.getProjectAreaAlias());
			
			//new paramters for proxy support
			if(null != server.getProxy())
				properties.setProperty(CommandLineParser.PROXY_PROPERTY_NAME, server.getProxy());
			if(null != server.getProxyPort())
				properties.setProperty(CommandLineParser.PROXY_PORT_PROPERTY_NAME, server.getProxyPort());
			if(null != server.getProxyUser())
				properties.setProperty(CommandLineParser.PROXY_USER_PROPERTY_NAME, server.getProxyUser());
			if(null != server.getProxyPassword())
				properties.setProperty(CommandLineParser.PROXY_PASSWORD_PROPERTY_NAME, server.getProxyPassword());
			
			fos = new FileOutputStream(configFile);
			properties.store(fos, COMMENT);
		} catch (FileNotFoundException e) {
			Logger.Log.error("storeProperties: Failed to save configuration file for the command line adapter");
			Logger.Log.debug(e);
		} catch (IOException e) {
			Logger.Log.error("storeProperties: Failed to save configuration file for the command line adapter");
			Logger.Log.debug(e);
		} finally {
			if (fos != null){
				try {
					fos.close();
				} catch (IOException e) {
					//do nothing
				}
			}
		}
	}
	
	private static boolean isRegisteredArgsChanged(Properties argsProps, Properties configFileProps)
	{
//		if(parameterExists(args, CommandLineParser.SLEEP_TIME_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.SLEEP_TIME_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.SLEEP_TIME_PROPERTY_NAME, configFile))
//		|| parameterExists(args, CommandLineParser.ADAPTER_ID_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.ADAPTER_ID_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.ADAPTER_ID_PROPERTY_NAME, configFile))
//		|| parameterExists(args, CommandLineParser.ADAPTER_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.ADAPTER_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.ADAPTER_PROPERTY_NAME, configFile))
//		)
//			return true;
//		else
//			return false;
		if (		
				argsProps.containsKey(CommandLineParser.SLEEP_TIME_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.SLEEP_TIME_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.SLEEP_TIME_PROPERTY_NAME)) ||
				argsProps.containsKey(CommandLineParser.ADAPTER_ID_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.ADAPTER_ID_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.ADAPTER_ID_PROPERTY_NAME)) ||
				argsProps.containsKey(CommandLineParser.ADAPTER_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.ADAPTER_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.ADAPTER_PROPERTY_NAME))){
			return true;
		} else {
			return false;
		}
	}
	
	private static Properties loadConfigFile(File configFile){
		Properties props = new Properties();
		if (configFile == null || !configFile.exists()){
			return props;
   		}
		FileInputStream propertyFile = null;
		try
		{
			propertyFile = new FileInputStream(configFile);
			if ( propertyFile != null )
			{
				props = new java.util.Properties() ;
				props.load(propertyFile) ;
			}
		} catch(Throwable e){
			Logger.Log.error("Error loading configuration file properties.");
		} finally {
			if (propertyFile != null){
				try {
					propertyFile.close();
				} catch (IOException e) {
					//do nothing;
				}
			}
		}
		
		return props;
	}
}
